@@ -19,7 +19,8 @@
 
 #include <nmofono/wifi/wifi-link-impl.h>
 #include <nmofono/wifi/access-point-impl.h>
-#include <nmofono/wifi/grouped-access-point-impl.h>
+#include <nmofono/wifi/grouped-access-point.h>
+#include <url-dispatcher-cpp/url-dispatcher.h>
 #include <cassert>
 
 #include <NetworkManagerActiveConnectionInterface.h>
@@ -28,18 +29,23 @@
 
 #include <NetworkManager.h>
 #include <iostream>
+#include <QUrlQuery>
 
-using namespace platform::nmofono::wifi;
-namespace networking = connectivity::networking;
+using namespace std;
 
-struct Link::Private: public QObject
+namespace nmofono
+{
+namespace wifi
+{
+
+struct WifiLinkImpl::Private: public QObject
 {
     Q_OBJECT
 
 public:
-    Private(Link& parent,
-            std::shared_ptr<OrgFreedesktopNetworkManagerDeviceInterface> dev,
-            std::shared_ptr<OrgFreedesktopNetworkManagerInterface> nm,
+    Private(WifiLinkImpl& parent,
+            shared_ptr<OrgFreedesktopNetworkManagerDeviceInterface> dev,
+            shared_ptr<OrgFreedesktopNetworkManagerInterface> nm,
             KillSwitch::Ptr killSwitch)
        : p(parent),
          m_dev(dev),
@@ -51,25 +57,29 @@ public:
     {
     }
 
-    Link& p;
+    WifiLinkImpl& p;
 
-    std::uint32_t m_characteristics = Link::Characteristics::empty;
-    connectivity::networking::Link::Status m_status = Status::disabled;
-    std::set<std::shared_ptr<connectivity::networking::wifi::AccessPoint>> m_rawAccessPoints;
-    std::set<std::shared_ptr<connectivity::networking::wifi::AccessPoint>> m_groupedAccessPoints;
-    std::shared_ptr<connectivity::networking::wifi::AccessPoint> m_activeAccessPoint;
+    uint32_t m_characteristics = Link::Characteristics::empty;
+    Link::Status m_status = Status::disabled;
+    QSet<AccessPointImpl::Ptr> m_rawAccessPoints;
+    QSet<AccessPoint::Ptr> m_groupedAccessPoints;
+    AccessPoint::Ptr m_activeAccessPoint;
+    Signal m_signal = Signal::disconnected;
 
-    std::shared_ptr<OrgFreedesktopNetworkManagerDeviceInterface> m_dev;
+    shared_ptr<OrgFreedesktopNetworkManagerDeviceInterface> m_dev;
     OrgFreedesktopNetworkManagerDeviceWirelessInterface m_wireless;
-    std::shared_ptr<OrgFreedesktopNetworkManagerInterface> m_nm;
+    shared_ptr<OrgFreedesktopNetworkManagerInterface> m_nm;
 
     KillSwitch::Ptr m_killSwitch;
 
-    std::map<AccessPoint::Key, std::shared_ptr<GroupedAccessPoint>> m_grouper;
-    std::uint32_t m_lastState = 0;
+    map<AccessPointImpl::Key, shared_ptr<GroupedAccessPoint>> m_grouper;
+    uint32_t m_lastState = 0;
     QString m_name;
-    std::shared_ptr<OrgFreedesktopNetworkManagerConnectionActiveInterface> m_activeConnection;
+    shared_ptr<OrgFreedesktopNetworkManagerConnectionActiveInterface> m_activeConnection;
+    unique_ptr<QMetaObject::Connection> m_signalStrengthConnection;
     bool m_connecting = false;
+    bool m_hideAccessPoints = false;
+    bool m_disconnectWifi = false;
 
     void setStatus(Status status)
     {
@@ -147,6 +157,16 @@ public:
 
     }
 
+
+    void disconnectSignalStengthConnection()
+    {
+        if (m_signalStrengthConnection)
+        {
+            disconnect(*m_signalStrengthConnection);
+            m_signalStrengthConnection.reset();
+        }
+    }
+
     /// '/' path means invalid.
     void updateActiveConnection(const QDBusObjectPath &path)
     {
@@ -155,6 +175,8 @@ public:
             m_activeAccessPoint.reset();
             Q_EMIT p.activeAccessPointUpdated(m_activeAccessPoint);
             m_activeConnection.reset();
+            disconnectSignalStengthConnection();
+            strengthUpdated();
             return;
         }
 
@@ -165,7 +187,7 @@ public:
         }
 
         try {
-            m_activeConnection = std::make_shared<
+            m_activeConnection = make_shared<
                     OrgFreedesktopNetworkManagerConnectionActiveInterface>(
                     NM_DBUS_SERVICE, path.path(), m_dev->connection());
             uint state = m_activeConnection->state();
@@ -180,15 +202,22 @@ public:
                 // for Wi-Fi devices specific_object is the AccessPoint object.
                 QDBusObjectPath ap_path = m_activeConnection->specificObject();
                 for (auto &ap : m_groupedAccessPoints) {
-                    auto shap =  std::dynamic_pointer_cast<GroupedAccessPoint>(ap);
+                    auto shap =  dynamic_pointer_cast<GroupedAccessPoint>(ap);
                     if (shap->has_object(ap_path)) {
                         m_activeAccessPoint = ap;
+                        disconnectSignalStengthConnection();
+                        m_signalStrengthConnection = make_unique<
+                                QMetaObject::Connection>(
+                                connect(m_activeAccessPoint.get(),
+                                        &AccessPoint::strengthUpdated, this,
+                                        &Private::strengthUpdated));
                         Q_EMIT p.activeAccessPointUpdated(m_activeAccessPoint);
+                        strengthUpdated();
                         break;
                     }
                 }
             }
-        } catch (std::exception &e) {
+        } catch (exception &e) {
             qWarning() << "failed to get active connection:";
             qWarning() << "\tpath: " << path.path();
             qWarning() << "\t" << QString::fromStdString(e.what());
@@ -201,19 +230,19 @@ public Q_SLOTS:
     {
         try {
             for (auto ap : m_rawAccessPoints) {
-                if (std::dynamic_pointer_cast<AccessPoint>(ap)->object_path() == path) {
+                if (dynamic_pointer_cast<AccessPoint>(ap)->object_path() == path) {
                     // already in the list
                     return;
                 }
             }
 
-            platform::nmofono::wifi::AccessPoint::Ptr shap;
+            AccessPointImpl::Ptr shap;
             try {
-                auto ap = std::make_shared<
+                auto ap = make_shared<
                         OrgFreedesktopNetworkManagerAccessPointInterface>(
                         NM_DBUS_SERVICE, path.path(), m_dev->connection());
-                shap = std::make_shared<platform::nmofono::wifi::AccessPoint>(ap);
-            } catch(const std::exception &e) {
+                shap = make_shared<AccessPointImpl>(ap);
+            } catch(const exception &e) {
                 qWarning() << __PRETTY_FUNCTION__ << ": failed to create AccessPoint proxy for "<< path.path() << ": ";
                 qWarning() << "\t" << QString::fromStdString(e.what());
                 qWarning() << "\tIgnoring.";
@@ -221,32 +250,31 @@ public Q_SLOTS:
             }
 
             m_rawAccessPoints.insert(shap);
-            Q_EMIT p.rawAccessPointsUpdated(m_rawAccessPoints);
 
-            auto k = AccessPoint::Key(shap);
+            auto k = AccessPointImpl::Key(shap);
             if(m_grouper.find(k) != m_grouper.end()) {
                 m_grouper[k]->add_ap(shap);
             } else {
-                m_grouper[k] = std::make_shared<GroupedAccessPoint>(shap);
+                m_grouper[k] = make_shared<GroupedAccessPoint>(shap);
             }
-            update_grouped();
-        } catch(const std::exception &e) {
+            update_grouped_access_points();
+        } catch(const exception &e) {
             /// @bug dbus-cpp internal logic exploded
             // If this happens, indicator-network is in an unknown state with no clear way of
             // recovering. The only reasonable way out is a graceful exit.
-            std::cerr << __PRETTY_FUNCTION__ << " Failed to run dbus service: " << e.what() << std::endl;
+            cerr << __PRETTY_FUNCTION__ << " Failed to run dbus service: " << e.what() << endl;
         }
     }
 
     void ap_removed(const QDBusObjectPath &path)
     {
-        platform::nmofono::wifi::AccessPoint::Ptr shap;
+        AccessPointImpl::Ptr shap;
 
         auto list = m_rawAccessPoints;
         for (const auto &ap : list) {
-            if (std::dynamic_pointer_cast<AccessPoint>(ap)->object_path() == path) {
-                shap = std::dynamic_pointer_cast<platform::nmofono::wifi::AccessPoint>(ap);
-                list.erase(ap);
+            if (ap->object_path() == path) {
+                shap = ap;
+                list.remove(ap);
                 break;
             }
         }
@@ -255,29 +283,36 @@ public Q_SLOTS:
             return;
         }
         m_rawAccessPoints = list;
-        Q_EMIT p.rawAccessPointsUpdated(m_rawAccessPoints);
 
-        for (auto &e : m_grouper) {
-            if (!e.second->has_object(path))
-                continue;
-
-            if(e.second->num_aps() == 1) {
-                m_grouper.erase(e.first);
-            } else {
-                e.second->remove_ap(shap);
+        AccessPointImpl::Key k(shap);
+        auto it = m_grouper.find(k);
+        if (it != m_grouper.end())
+        {
+            it->second->remove_ap(shap);
+            if (it->second->num_aps() == 0)
+            {
+                m_grouper.erase(it);
             }
-            break;
         }
-        update_grouped();
+        update_grouped_access_points();
     }
 
-    void update_grouped() {
-        std::set<std::shared_ptr<connectivity::networking::wifi::AccessPoint>> new_grouped;
-        for(auto &i : m_grouper) {
-            new_grouped.insert(i.second);
+    void update_grouped_access_points()
+    {
+        m_groupedAccessPoints.clear();
+        for (auto &i : m_grouper)
+        {
+            m_groupedAccessPoints.insert(i.second);
+        }
+
+        if (m_hideAccessPoints)
+        {
+            Q_EMIT p.accessPointsUpdated(QSet<AccessPoint::Ptr>());
+        }
+        else
+        {
+            Q_EMIT p.accessPointsUpdated(m_groupedAccessPoints);
         }
-        m_groupedAccessPoints = new_grouped;
-        Q_EMIT p.accessPointsUpdated(m_groupedAccessPoints);
     }
 
     void state_changed(uint new_state, uint, uint)
@@ -285,15 +320,55 @@ public Q_SLOTS:
         updateDeviceState(new_state);
     }
 
-    void kill_switch_updated()
+    void kill_switch_updated(KillSwitch::State)
     {
         updateDeviceState(m_lastState);
     }
+
+    void strengthUpdated()
+    {
+        Signal signal = Signal::disconnected;
+
+        if (m_activeAccessPoint && !m_hideAccessPoints)
+        {
+            double strength = m_activeAccessPoint->strength();
+            bool secured  = m_activeAccessPoint->secured();
+
+            if (strength >= 80.0)
+            {
+                signal = secured ? Signal::signal_100_secure : Signal::signal_100;
+            }
+            else if (strength >= 60.0)
+            {
+                signal = secured ? Signal::signal_75_secure : Signal::signal_75;
+            }
+            else if (strength >= 40.0)
+            {
+                signal = secured ? Signal::signal_50_secure : Signal::signal_50;
+            }
+            else if (strength >= 20.0)
+            {
+                signal = secured ? Signal::signal_25_secure : Signal::signal_25;
+            }
+            else
+            {
+                signal = secured ? Signal::signal_0_secure : Signal::signal_0;
+            }
+        }
+
+        if (m_signal == signal)
+        {
+            return;
+        }
+
+        m_signal = signal;
+        Q_EMIT p.signalUpdated(m_signal);
+    }
 };
 
-Link::Link(std::shared_ptr<OrgFreedesktopNetworkManagerDeviceInterface> dev,
-           std::shared_ptr<OrgFreedesktopNetworkManagerInterface> nm,
-           platform::nmofono::KillSwitch::Ptr killSwitch)
+WifiLinkImpl::WifiLinkImpl(shared_ptr<OrgFreedesktopNetworkManagerDeviceInterface> dev,
+           shared_ptr<OrgFreedesktopNetworkManagerInterface> nm,
+           KillSwitch::Ptr killSwitch)
     : d(new Private(*this, dev, nm, killSwitch)) {
     d->m_name = d->m_dev->interface();
 
@@ -308,140 +383,81 @@ Link::Link(std::shared_ptr<OrgFreedesktopNetworkManagerDeviceInterface> dev,
     d->updateDeviceState(d->m_dev->state());
 
     connect(d->m_killSwitch.get(), &KillSwitch::stateChanged, d.get(), &Private::kill_switch_updated);
+
+    d->strengthUpdated();
 }
 
-Link::~Link()
+WifiLinkImpl::~WifiLinkImpl()
 {}
 
-void
-Link::enable()
-{
-#ifdef INDICATOR_NETWORK_TRACE_MESSAGES
-    qDebug() << "Link::enable()";
-#endif
-
-    try {
-        if (d->m_killSwitch->state() != KillSwitch::State::unblocked) {
-            // try to unblock. throws if fails.
-            d->m_killSwitch->unblock();
-        }
-        d->m_nm->setWirelessEnabled(true);
-        d->m_dev->setAutoconnect(true);
-    } catch(std::runtime_error &e) {
-        /// @todo when toggling enable()/disable() rapidly the default timeout of
-        ///       1 second in dbus::core::Property is not long enough..
-        ///       just ignore for now and get dbus-cpp to have larger timeout.
-        std::cerr << __PRETTY_FUNCTION__ << ": " << e.what() << std::endl;
-    }
-
-#if 0
-    /** @todo:
-     * NetworkManager 0.9.9 allows doing this below and letting NM to figure out the best connection:
-     *
-     * m_nm.activate_connection(dbus::types::ObjectPath("/"),
-     *                          m_dev.object->path(),
-     *                          core::dbus::typ es::ObjectPath("/"));
-     *
-     * for now we get somewhat similar result by toggling the Autoconnect property,
-     * but this breaks the possible setting we will have for autoconnecting later.
-     */
-    m_dev.autoconnect->set(true);
-    m_disabled = false;
-#endif
-}
-
-void
-Link::disable()
-{
-#ifdef INDICATOR_NETWORK_TRACE_MESSAGES
-    std::cout << "Link::disable()" << std::endl;
-#endif
-
-    /// @todo for now just disable wireless completely.
-    ///        this only works properly when there is one wifi adapter on the system
-    try {
-        if (d->m_killSwitch->state() == KillSwitch::State::unblocked) {
-            // block the device. that will disable it also
-            d->m_killSwitch->block();
-            return;
-        }
-        d->m_nm->setWirelessEnabled(false);
-        d->m_dev->setAutoconnect(false);
-    } catch(std::runtime_error &e) {
-        /// @todo when toggling enable()/disable() rapidly the default timeout of
-        ///       1 second in dbus::core::Property is not long enough..
-        ///       just ignore for now and get dbus-cpp to have larger timeout.
-        qWarning() << __PRETTY_FUNCTION__ << ": " << QString::fromStdString(e.what());
-    }
-
-#if 0
-    /** @todo remove this after NM 0.9.9 is available. check comment in enable() */
-    m_dev.disconnect();
-#endif
-}
-
-networking::Link::Type
-Link::type() const
+Link::Type
+WifiLinkImpl::type() const
 {
     return Type::wifi;
 }
 
 uint32_t
-Link::characteristics() const
+WifiLinkImpl::characteristics() const
 {
     return d->m_characteristics;
 }
 
-networking::Link::Status
-Link::status() const
+Link::Status
+WifiLinkImpl::status() const
 {
     return d->m_status;
 }
 
-networking::Link::Id
-Link::id() const
+Link::Id
+WifiLinkImpl::id() const
 {
     return 0;
 }
 
 QString
-Link::name() const
+WifiLinkImpl::name() const
 {
     return d->m_name;
 }
 
-const std::set<std::shared_ptr<networking::wifi::AccessPoint> >&
-Link::rawAccessPoints() const
+WifiLink::Mode WifiLinkImpl::mode() const
 {
-    return d->m_rawAccessPoints;
+    static const map<uint, WifiLink::Mode> modeMap
+    {
+        {NM_802_11_MODE_UNKNOWN, Mode::unknown},
+        {NM_802_11_MODE_ADHOC, Mode::adhoc},
+        {NM_802_11_MODE_INFRA, Mode::infra},
+        {NM_802_11_MODE_AP, Mode::ap}
+    };
+    return modeMap.find(d->m_wireless.mode())->second;
 }
 
-const std::set<std::shared_ptr<connectivity::networking::wifi::AccessPoint>>&
-Link::accessPoints() const {
+WifiLink::Signal WifiLinkImpl::signal() const
+{
+    return d->m_signal;
+}
+
+QSet<AccessPoint::Ptr>
+WifiLinkImpl::accessPoints() const {
+    if (d->m_hideAccessPoints)
+    {
+        return QSet<AccessPoint::Ptr>();
+    }
+
     return d->m_groupedAccessPoints;
 }
 
 void
-Link::connect_to(std::shared_ptr<networking::wifi::AccessPoint> accessPoint)
+WifiLinkImpl::connect_to(AccessPoint::Ptr accessPoint)
 {
     try {
         d->m_connecting = true;
-        QByteArray ssid;
-        std::shared_ptr<GroupedAccessPoint> ap = std::dynamic_pointer_cast<GroupedAccessPoint>(accessPoint);
-        // The accesspoint interface does not provide this property so we need to coax it out of
-        // derived classes.
-        if(ap) {
-            ssid = ap->raw_ssid();
-        } else {
-            std::shared_ptr<AccessPoint> bap = std::dynamic_pointer_cast<AccessPoint>(accessPoint);
-            assert(bap);
-            ssid = bap->raw_ssid();
-        }
+        QByteArray ssid = accessPoint->raw_ssid();
 
-        std::shared_ptr<OrgFreedesktopNetworkManagerSettingsConnectionInterface> found;
+        shared_ptr<OrgFreedesktopNetworkManagerSettingsConnectionInterface> found;
         QList<QDBusObjectPath> connections = d->m_dev->availableConnections();
         for (auto &path : connections) {
-            auto con = std::make_shared<OrgFreedesktopNetworkManagerSettingsConnectionInterface>(
+            auto con = make_shared<OrgFreedesktopNetworkManagerSettingsConnectionInterface>(
                     NM_DBUS_SERVICE, path.path(), d->m_dev->connection());
             QVariantDictMap settings = con->GetSettings();
             auto wirelessIt = settings.find("802-11-wireless");
@@ -450,8 +466,7 @@ Link::connect_to(std::shared_ptr<networking::wifi::AccessPoint> accessPoint)
                 auto ssidIt = wirelessIt->find("ssid");
                 if (ssidIt != wirelessIt->cend())
                 {
-                    QByteArray value = ssidIt->toByteArray();
-                    if (value == ap->raw_ssid())
+                    if (ssidIt->toByteArray() == ssid)
                     {
                         found = con;
                         break;
@@ -467,39 +482,89 @@ Link::connect_to(std::shared_ptr<networking::wifi::AccessPoint> accessPoint)
         if (found) {
             ac = d->m_nm->ActivateConnection(QDBusObjectPath(found->path()),
                                            QDBusObjectPath(d->m_dev->path()),
-                    ap->object_path());
+                                           accessPoint->object_path());
         } else {
-            QVariantDictMap conf;
+            if (accessPoint->enterprise()) {
+                // activate system settings URI
+                QUrlQuery q;
+                q.addQueryItem("ssid", accessPoint->raw_ssid());
+                q.addQueryItem("bssid", accessPoint->bssid());
+                QString url = "settings:///system/wifi?" + q.query(QUrl::FullyEncoded);
 
-            /// @todo getting the ssid multiple times over dbus is stupid.
+                UrlDispatcher::send(url.toStdString(), [](string url, bool success) {
+                    if (!success) {
+                        cerr << "URL Dispatcher failed on " << url << endl;
+                    }
+                });
+            } else {
+                QVariantDictMap conf;
 
-            QVariantMap wireless_conf;
-            wireless_conf["ssid"] = ap->raw_ssid();
+                /// @todo getting the ssid multiple times over dbus is stupid.
 
-            conf["802-11-wireless"] = wireless_conf;
-            auto ret = d->m_nm->AddAndActivateConnection(
-                    conf, QDBusObjectPath(d->m_dev->path()), ap->object_path());
-            ret.waitForFinished();
-            ac = ret.argumentAt<1>();
+                QVariantMap wireless_conf;
+                wireless_conf["ssid"] = ssid;
+
+                conf["802-11-wireless"] = wireless_conf;
+                auto ret = d->m_nm->AddAndActivateConnection(
+                        conf, QDBusObjectPath(d->m_dev->path()), accessPoint->object_path());
+                ret.waitForFinished();
+                ac = ret.argumentAt<1>();
+            }
+        }
+        // For enterprise access points, the system settings app will perform the connection
+        if (!accessPoint->enterprise()) {
+            d->updateActiveConnection(ac);
         }
-        d->updateActiveConnection(ac);
         d->m_connecting = false;
-    } catch(const std::exception &e) {
+    } catch(const exception &e) {
         // @bug default timeout expired: LP(#1361642)
         // If this happens, indicator-network is in an unknown state with no clear way of
         // recovering. The only reasonable way out is a graceful exit.
-        std::cerr << __PRETTY_FUNCTION__ << " Failed to activate connection: " << e.what() << std::endl;
+        qWarning() << __PRETTY_FUNCTION__ << " Failed to activate connection: " << e.what();
     }
 }
 
-std::shared_ptr<networking::wifi::AccessPoint>
-Link::activeAccessPoint()
+AccessPoint::Ptr
+WifiLinkImpl::activeAccessPoint()
 {
     return d->m_activeAccessPoint;
 }
 
-QDBusObjectPath Link::device_path() const {
+QDBusObjectPath WifiLinkImpl::device_path() const {
     return QDBusObjectPath(d->m_dev->path());
 }
 
+void
+WifiLinkImpl::setHideAccessPoints(bool hide)
+{
+    if (hide == d->m_hideAccessPoints)
+    {
+        return;
+    }
+
+    d->m_hideAccessPoints = hide;
+    d->update_grouped_access_points();
+    d->strengthUpdated();
+}
+
+void
+WifiLinkImpl::setDisconnectWifi(bool disconnect)
+{
+    if (disconnect == d->m_disconnectWifi)
+    {
+        return;
+    }
+
+    d->m_dev->setAutoconnect(!disconnect);
+    if (disconnect && d->m_activeConnection)
+    {
+        // Disconnect from the current network
+        d->m_nm->DeactivateConnection(
+                QDBusObjectPath(d->m_activeConnection->path()));
+    }
+}
+
+}
+}
+
 #include "wifi-link-impl.moc"
